#include <cuda_runtime.h>
#include <device_launch_parameters.h>
#include <device_functions.h>

#include <cstdio>
#include <cstdlib>

const int N = 256*512;
const int BLOCK_SIZE = 256;
const float EPS = 0.0001f;

/*----------------------------------------------------------------------------------------------------------*/
__global__ void kernelFloat3(float3 *newPos, float3 * newVel, float3 *oldPos, float3 *oldVel, float dt){
	int index = blockIdx.x*blockDim.x + threadIdx.x;

	float3 pos = oldPos[index];
	float3 f = make_float3(0.0f, 0.0f, 0.0f);

	for(int i = 0; i < N; ++i){
		float3 pi = oldPos[i];
		float3 r = make_float3(pi.x - pos.x, pi.y - pos.y, pi.z - pos.z);

		float invDist = 1.0f/sqrtf(r.x*r.x + r.y*r.y + r.z*r.z + EPS*EPS);
		float s = invDist*invDist*invDist;

		f.x += r.x*s;
		f.y += r.y*s;
		f.z += r.z*s;
	}

	float3 vel = oldVel[index];

	vel.x += f.x*dt;
	vel.y += f.y*dt;
	vel.z += f.z*dt;

	pos.x += vel.x*dt;
	pos.y += vel.y*dt;
	pos.z += vel.z*dt;

	newPos[index] = pos;
	newVel[index] = vel;
}

float solveFloat3(){
	float3 *p = new float3[N];
	float3 *v = new float3[N];
	float3 *pDev[2] = {NULL, NULL};
	float3 *vDev[2] = {NULL, NULL};

	for(int i = 0; i < N; ++i){
		p[i] = make_float3(rand()/(float)RAND_MAX - 0.5f, rand()/(float)RAND_MAX - 0.5f,rand()/(float)RAND_MAX - 0.5f);
		v[i] = make_float3(rand()/(float)RAND_MAX - 0.5f, rand()/(float)RAND_MAX - 0.5f,rand()/(float)RAND_MAX - 0.5f);
	}

	cudaMalloc((void **)&pDev[0], N*sizeof(float3));
	cudaMalloc((void **)&pDev[1], N*sizeof(float3));
	cudaMalloc((void **)&vDev[0], N*sizeof(float3));
	cudaMalloc((void **)&vDev[1], N*sizeof(float3));
	cudaMemcpy(pDev[0], p, N*sizeof(float3), cudaMemcpyHostToDevice);
	cudaMemcpy(vDev[0], v, N*sizeof(float3), cudaMemcpyHostToDevice);

	cudaEvent_t start, stop;
	float gpuTime = 0.0f;

	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	cudaEventRecord(start, 0);

	int index = 0;
	for(int i = 0; i < 2; ++i, index ^= 1){
		kernelFloat3<<<dim3(N/BLOCK_SIZE), dim3(BLOCK_SIZE)>>>(pDev[index^1], vDev[index^1],
															   pDev[index], vDev[index], 0.01f);
	}
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&gpuTime, start, stop);

	cudaMemcpy(p, pDev[index^1], N*sizeof(float3), cudaMemcpyDeviceToHost);
	cudaMemcpy(v, vDev[index^1], N*sizeof(float3), cudaMemcpyDeviceToHost);

	cudaFree(pDev[0]);
	cudaFree(pDev[1]);
	cudaFree(vDev[0]);
	cudaFree(vDev[1]);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);
	cudaDeviceReset();

	delete[] p;
	delete[] v;
	return gpuTime;
}
/*----------------------------------------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------------------------------------*/
__global__ void kernelFloat4(float4 *newPos, float4 * newVel, float4 *oldPos, float4 *oldVel, float dt){
	int index = blockIdx.x*blockDim.x + threadIdx.x;

	float4 pos = oldPos[index];
	float3 f = make_float3(0.0f, 0.0f, 0.0f);

	for(int i = 0; i < N; ++i){
		float4 pi = oldPos[i];
		float3 r = make_float3(pi.x - pos.x, pi.y - pos.y, pi.z - pos.z);

		float invDist = 1.0f/sqrtf(r.x*r.x + r.y*r.y + r.z*r.z + EPS*EPS);
		float s = invDist*invDist*invDist;

		f.x += r.x*s;
		f.y += r.y*s;
		f.z += r.z*s;
	}

	float4 vel = oldVel[index];

	vel.x += f.x*dt;
	vel.y += f.y*dt;
	vel.z += f.z*dt;

	pos.x += vel.x*dt;
	pos.y += vel.y*dt;
	pos.z += vel.z*dt;

	newPos[index] = pos;
	newVel[index] = vel;
}

float solveFloat4(){
	float4 *p = new float4[N];
	float4 *v = new float4[N];
	float4 *pDev[2] = {NULL, NULL};
	float4 *vDev[2] = {NULL, NULL};

	for(int i = 0; i < N; ++i){
		p[i] = make_float4(rand()/(float)RAND_MAX - 0.5f, rand()/(float)RAND_MAX - 0.5f,rand()/(float)RAND_MAX - 0.5f, 0.0f);
		v[i] = make_float4(rand()/(float)RAND_MAX - 0.5f, rand()/(float)RAND_MAX - 0.5f,rand()/(float)RAND_MAX - 0.5f, 0.0f);
	}

	cudaMalloc((void **)&pDev[0], N*sizeof(float4));
	cudaMalloc((void **)&pDev[1], N*sizeof(float4));
	cudaMalloc((void **)&vDev[0], N*sizeof(float4));
	cudaMalloc((void **)&vDev[1], N*sizeof(float4));
	cudaMemcpy(pDev[0], p, N*sizeof(float4), cudaMemcpyHostToDevice);
	cudaMemcpy(vDev[0], v, N*sizeof(float4), cudaMemcpyHostToDevice);

	cudaEvent_t start, stop;
	float gpuTime = 0.0f;

	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	cudaEventRecord(start, 0);

	int index = 0;
	for(int i = 0; i < 2; ++i, index ^= 1){
		kernelFloat4<<<dim3(N/BLOCK_SIZE), dim3(BLOCK_SIZE)>>>(pDev[index^1], vDev[index^1],
															   pDev[index], vDev[index], 0.01f);
	}

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&gpuTime, start, stop);

	cudaMemcpy(p, pDev[index^1], N*sizeof(float4), cudaMemcpyDeviceToHost);
	cudaMemcpy(v, vDev[index^1], N*sizeof(float4), cudaMemcpyDeviceToHost);

	cudaFree(pDev[0]);
	cudaFree(pDev[1]);
	cudaFree(vDev[0]);
	cudaFree(vDev[1]);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);
	cudaDeviceReset();

	delete[] p;
	delete[] v;
	return gpuTime;
}
/*----------------------------------------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------------------------------------*/
__global__ void kernelFloat4Shared(float4 *newPos, float4 * newVel, float4 *oldPos, float4 *oldVel, float dt){
	int index = blockIdx.x*blockDim.x + threadIdx.x;

	float4 pos = oldPos[index];
	float3 f = make_float3(0.0f, 0.0f, 0.0f);
	int ind = 0;

	__shared__ float4 sp[BLOCK_SIZE];

	for(int i = 0; i < N/BLOCK_SIZE; ++i, ind += BLOCK_SIZE){
		sp[threadIdx.x] = oldPos[ind + threadIdx.x];
		__syncthreads();

		for(int j = 0; j < 256; ++j){
			float3 r;

			r.x = sp[j].x - pos.x;
			r.y = sp[j].y - pos.y;
			r.z = sp[j].z - pos.z;

			float invDist = 1.0f/sqrtf(r.x*r.x + r.y*r.y + r.z*r.z + EPS*EPS);
			float s = invDist*invDist*invDist;

			f.x += r.x*s;
			f.y += r.y*s;
			f.z += r.z*s;
		}

		__syncthreads();
	}

	float4 vel = oldVel[index];

	vel.x += f.x*dt;
	vel.y += f.y*dt;
	vel.z += f.z*dt;

	pos.x += vel.x*dt;
	pos.y += vel.y*dt;
	pos.z += vel.z*dt;

	newPos[index] = pos;
	newVel[index] = vel;
}

float solveFloat4Shared(){
	float4 *p = new float4[N];
	float4 *v = new float4[N];
	float4 *pDev[2] = {NULL, NULL};
	float4 *vDev[2] = {NULL, NULL};

	for(int i = 0; i < N; ++i){
		p[i] = make_float4(rand()/(float)RAND_MAX - 0.5f, rand()/(float)RAND_MAX - 0.5f,rand()/(float)RAND_MAX - 0.5f, 0.0f);
		v[i] = make_float4(rand()/(float)RAND_MAX - 0.5f, rand()/(float)RAND_MAX - 0.5f,rand()/(float)RAND_MAX - 0.5f, 0.0f);
	}

	cudaMalloc((void **)&pDev[0], N*sizeof(float4));
	cudaMalloc((void **)&pDev[1], N*sizeof(float4));
	cudaMalloc((void **)&vDev[0], N*sizeof(float4));
	cudaMalloc((void **)&vDev[1], N*sizeof(float4));
	cudaMemcpy(pDev[0], p, N*sizeof(float4), cudaMemcpyHostToDevice);
	cudaMemcpy(vDev[0], v, N*sizeof(float4), cudaMemcpyHostToDevice);

	cudaEvent_t start, stop;
	float gpuTime = 0.0f;

	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	cudaEventRecord(start, 0);

	int index = 0;
	for(int i = 0; i < 2; ++i, index ^= 1){
		kernelFloat4Shared<<<dim3(N/BLOCK_SIZE), dim3(BLOCK_SIZE)>>>(pDev[index^1], vDev[index^1],
															   pDev[index], vDev[index], 0.01f);
	}

	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&gpuTime, start, stop);

	cudaMemcpy(p, pDev[index^1], N*sizeof(float4), cudaMemcpyDeviceToHost);
	cudaMemcpy(v, vDev[index^1], N*sizeof(float4), cudaMemcpyDeviceToHost);

	cudaFree(pDev[0]);
	cudaFree(pDev[1]);
	cudaFree(vDev[0]);
	cudaFree(vDev[1]);

	cudaEventDestroy(start);
	cudaEventDestroy(stop);
	cudaDeviceReset();

	delete[] p;
	delete[] v;
	return gpuTime;
}

/*----------------------------------------------------------------------------------------------------------*/

int main(int argc, char *argv[]){
	float FLOAT3 = solveFloat3();
	float FLOAT4 = solveFloat4();
	float FLOAT4Shared = solveFloat4Shared();

	printf("float3: %.7f\n", FLOAT3);
	printf("float4: %.7f\n", FLOAT4);
	printf("float4shared: %.7f\n", FLOAT4Shared);
	return 0;
}
